#ifndef QUATERNION_H
#define QUATERNION_H

#include "Vector.h"

class Point;
class Matrix;

// TODO: derive from Tuple<4>

class Quaternion
{
public:
	union
	{
		struct
		{
			float x;
			float y;
			float z;
		};
		struct
		{
			Vector v;
		};
	};
	float w;

	Quaternion(){};
	Quaternion(float w);
	Quaternion(const Vector &v);
	Quaternion(const Quaternion &Q);
	Quaternion(const Quaternion &Q, const Quaternion &R, float t);   // Slerp
	Quaternion(const Vector &v, float w = 0.0f);
	Quaternion(const Point &v, float w = 1.0f);
	Quaternion(const Matrix &A);   // Must be a rotation Matrix
	Quaternion(float Qx, float Qy, float Qz, float Qw);

	Quaternion &operator=(const Quaternion &Q);

	operator float() const;
	operator Vector() const;
	operator Matrix() const;

	Quaternion operator+() const;
	Quaternion operator-() const;

	Quaternion operator!() const;   // Inverse
	Quaternion operator~() const;   // Conjugate

	Quaternion &operator+=(const Quaternion &V);
	Quaternion &operator-=(const Quaternion &V);
	Quaternion &operator*=(float s);
	Quaternion &operator*=(const Quaternion &Q);
	Quaternion &operator/=(float s);

	friend bool operator==(const Quaternion &Q, const Quaternion &R);
	friend bool operator!=(const Quaternion &Q, const Quaternion &R);

	friend Quaternion operator+(const Quaternion &Q, const Quaternion &R);
	friend Quaternion operator-(const Quaternion &Q, const Quaternion &R);
	friend Quaternion operator*(const Quaternion &Q, const Quaternion &R);
	friend Quaternion operator*(float s, const Quaternion &Q);
	friend Quaternion operator*(const Quaternion &Q, float s);
	friend Quaternion operator/(const Quaternion &Q, float s);

	friend float N(const Quaternion &Q);   // Norm
	friend Quaternion slerp(const Quaternion &Q, const Quaternion &R, float t);

	Quaternion &normalize();
};

const Quaternion i(1, 0, 0, 0);   // Imaginary unit
const Quaternion j(0, 1, 0, 0);
const Quaternion k(0, 0, 1, 0);

inline Quaternion slerp(const Quaternion &Q, const Quaternion &R, float t)
{
	return Quaternion(Q, R, t);
}

#endif   // QUATERNION_H