#include "Render.h"

#include "Vector.h"
#include "Matrix.h"
#include "Quaternion.h"
#include "RandVec.h"
#include "DDDisplay.h"
#include "DIKeyboard.h"
#include "DIMouse.h"
#include "FPSCounter.h"

#include <stdlib.h>
#include <math.h>

#define NUM_RAND_POINTS 8192

void Render()
{
	static FPSCounter FPSCounter;

	static Vector *v = NULL;

	if(v == NULL)
	{
		Display.ClearBack();
		Display.Text(0, 0, " Loading... ");
		Display.Flip();

		v = new Vector[NUM_RAND_POINTS];

		for(unsigned int i = 0; i < NUM_RAND_POINTS; i++)
			v[i] = RandVec();
	}

	static Quaternion View = 1;

	Matrix M = (Matrix)View;

	Display.ClearBack();
	Display.Lock();

	const float W = 0.5f * Display.Width;
	const float H = 0.5f * Display.Height;
	const float D = 0.5f * Display.Width;   // Distance to projection plane, 90 FOV
	unsigned int *DisplayAddress = (unsigned int*)Display.BackBuffer;

	for(unsigned int i = 0; i < NUM_RAND_POINTS; i++)
	{
		Vector u = M * v[i];

		if(u.y > 0.01f)   // Not behind view
		{
			float Z = 1.0f / u.y;

			u.x = W + D * u.x * Z;
			u.y = H - D * u.z * Z;

			unsigned int X = (unsigned int)u.x;
			unsigned int Y = (unsigned int)u.y;

			if(X < Display.Width && Y < Display.Height)
				*(DisplayAddress + Display.Stride * Y + X) = 0xFFFFFFFF;
		}
	}

	Display.Unlock();
	Display.Text(0, 0, " %.1f FPS ", FPSCounter.CurrentFPS);
	Display.Flip();

	Keyboard.Input();

	if(Keyboard.KeyPressed(DIK_ESCAPE))
	{
		delete[] v;
		exit(0);
	}

	Mouse.Input();

	float SinPhi = sinf(1e-3f * Mouse.dx);
	float SinTheta = sinf(1e-3f * Mouse.dy);
	View = Quaternion(SinTheta, 0, SinPhi, sqrtf(1 - SinPhi*SinPhi - SinTheta*SinTheta)) * View;

	View.normalize();

	FPSCounter.NewFrame();
}
